/*
 * OctoMap - An Efficient Probabilistic 3D Mapping Framework Based on Octrees
 * http://octomap.github.com/
 *
 * Copyright (c) 2009-2013, K.M. Wurm and A. Hornung, University of Freiburg
 * All rights reserved.
 * License: New BSD
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of Freiburg nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <octomap/octomap.h>
#include <string.h>
#include <stdlib.h>

using namespace std;
using namespace octomap;

void printUsage(char* self){
  std::cerr << "\nUSAGE: " << self << " InputFile.log OutputFile.graph\n\n";

  std::cerr << "This tool converts a plain text log file into a binary scangraph file" << std::endl;
  std::cerr << "which can be used in Octomap.\n\n";
  std::cerr << "The log file needs to be in the format of:\n"
      << "NODE x y z roll pitch yaw\n"
      << "x y z\nx y z\n...\n"
      << "NODE x y z roll pitch yaw\n"
      << "x y z\n...\n\n"
      << "Lines starting with '#' or empty lines are ignored.\n\n";

  exit(0);
}

int main(int argc, char** argv) {
  // default values:
  string logFilename = "";
  string graphFilename = "";


  if (argc != 3){
    printUsage(argv[0]);
  } else{
    logFilename = std::string(argv[1]);
    graphFilename = std::string(argv[2]);
  }

  cout << "\nReading Log file\n===========================\n";
  ScanGraph* graph = new ScanGraph();
  graph->readPlainASCII(logFilename);

  cout << "\nWriting binary graph file\n===========================\n";

  graph->writeBinary(graphFilename);
  
  return 0;
}
