
/**
 * \namespace dynamicEDT3D Namespace of the dynamicEDT3D library
 *
 */


/** \mainpage dynamicEDT3D

  \section intro_sec Introduction

  The <a href="http://octomap.sourceforge.net/">dynamicEDT3D library</a> implements an inrementally updatable Euclidean distance transform (EDT) in 3D. It comes with a wrapper to use the OctoMap 3D representation and hooks into the change detection of the OctoMap library to propagate changes to the EDT.
  
  \section install_sec Installation
  <p>See the file README.txt in the main folder.
  </p>

  \section gettingstarted_sec Getting Started
  <p>There are two example programs in the src/examples directory that show the basic functionality of the library.</p>


  \section changelog_sec Changelog
  <p>See the file CHANGELOG.txt in the main folder or the <a href="http://octomap.svn.sourceforge.net/viewvc/octomap/trunk/dynamicEDT3D/CHANGELOG.txt">latest version online</a>
  </p>

  \section using_sec Using dynamicEDT3D?
  <p> Please let us know if you are using dynamicEDT3D, as we are curious to find out how it enables other people's work or research. Additionally, please cite our paper if you use dynamicEDT3D in your research: </p>
  <p>B. Lau, C. Sprunk, and W. Burgard, 
  <strong>"Efficient Grid-based Spatial Representations for Robot Navigation in Dynamic Environments"</strong> in
<em>Robotics and Autonomous Systems</em>, 2012. Accepted for publication.
Software available at <a href="http://octomap.sf.net/">http://octomap.sf.net/</a>.
</p>
<p>BibTeX:</p>
<pre>@@article{lau12ras,
  author = {Boris Lau and Christoph Sprunk and Wolfram Burgard},
  title = {Efficient Grid-based Spatial Representations for Robot Navigation in Dynamic Environments},
  journal = {Robotics and Autonomous Systems (RAS)},
  year = {2012},
  url = {http://octomap.sf.net/},
  note = {Accepted for publication. Software available at \url{http://octomap.sf.net/}}
}</pre>

**/




